// AGI Stack Patch

#include <stdio.h>
#include <stdlib.h>
#include "types.h"

#define AGI_FILENAME "agi.exe"
#define VERSION "1.0"
#define AGI_MIN 34300
#define AGI_MAX 40600
#define STACK_MIN 0x0A00
#define STACK_MAX 0xFFFF

/*
todo:
check the current stack is ok.. compare them all
print out the current stack
print default stack

notes:  adecrypt requires sierra.com
*/

/*
	SEARCH PATTERNS.........
	s = search string,     size = size in bytes,       m = mask where... 
	1 = check, 0=ignore, 2=ignore, but return address of first occurance if only occurs once
*/

struct patch_struct
{
	u8 *s;
	u8 *m;
	int size;
};
typedef struct patch_struct PATCH;

// first occurance of stack size:
PATCH patch1 = {	"\xE8\x00\x00\xC7\x46\xFE\x02\x02\xFF\x76\xFE" ,
				"10011122111" , 11  };
// second
PATCH patch2 = {"\x8B\xFE\xB9\x02\x02\xF3\xAA", "1112211", 7};
// third
PATCH patch3 = {	"\xF3\xAA\xC7\x04\xAA\xAA\x81\xC6\x02\x02\xFA",
				"11111111221", 11};
// fourth and final: (stack interupt)
PATCH patch4 = {	"\x8E\xD8\x8E\xC0\xA1\x00\x00\x05\x02\x02\x8B\xE0",
				"111111112211", 12};

/*
	CODE............
*/
	
void my_error(char *s)
{
	printf("%s", s);
	// do some other stuff that's error related.. like dance or something
	exit(1);
}				
				
				
u8 *search_data(u8 *d, int dsize, PATCH patch)
{
	int i, j;
	int count = 0;
	int mid_count = 0;
	u8 *p = 0;
	u8 *temp_p = 0;
	
	for (i=0; i<(dsize-patch.size) ; i++)
	{
		j = 0;
		mid_count = 0;
		//printf("0x%X ", i);
		while (    ( (d[i+j]==patch.s[j]) || (patch.m[j] != '1') ) && (j<patch.size)    ) 
		{
			if ( (patch.m[j] == '2') && (mid_count == 0) )
			{
				temp_p = d + i+j;
				mid_count = 1;
			}
			j++;
		}
		if (j==patch.size)
		{
			if (count == 0) p = temp_p;
			count ++;
		}
	}
	
	if (count == 0)
		my_error("no pattern found.  cannot reliably edit this file.\n");
	
	if (count != 0) printf("found (%d) ", count);
	return p;
}
	







int main (int argc, char *argv[])
{
	u8 *agi_data;
	u8 *p1, *p2, *p3, *p4;
	fpos_t agi_size;
	FILE *agi_file;
	
	u8 new_value[2];
	
	u8 *filename;
	long int newstack;
	
	// for the console window thingy
	//freopen( "CON", "w", stdout );
	
	// obligatory introduction
	printf("AGI Stack Patch v%s\n", VERSION);
	printf("Copyright (C) February 2001 Nick Sonneveld\n");
	printf("Compiled at %s , %s\n\n", __TIME__, __DATE__);
	
	printf("Based upon the Adventure Game Interpreter (AGI)\n");
	printf("Copyright (C) 1984, 1985, 1986 Sierra On-Line, Inc.\n");
	printf("Authors: Jeff Stephenson & Chris Iden\n\n\n");


	// read the command line parameters (name, new size)
	if (argc < 3)
		my_error("not enough parameters\n");
	
	filename = argv[1];
	newstack = strtol(argv[2], 0, 0);
	printf("New stack sized required = 0x%04lX\n", newstack);
	if ( (newstack < STACK_MIN) || (newstack > STACK_MAX) )
		my_error("stack size must be between 0x0A00 and 0xFFFF\n"  );
	
	
	// open the file
	printf("Opening AGI...\n");
	agi_file = fopen(filename, "r+b");
	
	/* The ISO standard says that when using such a stream, 
	you must call fflush (see section Stream Buffering) or a 
	file positioning function such as fseek (see section File 
	Positioning) when switching from reading to writing or vice versa.*/
	
	if (agi_file == 0)
		my_error("error opening the agi file.  does it exist?\n");
	
	
	// get the file size
	printf("Reading into memory...\n");	
	
	if (fseek(agi_file, 0, SEEK_END) != 0)
		my_error("error seeking to end of agi file\n");
	if (fgetpos(agi_file, &agi_size) != 0)
		my_error("error reading the file position\n");
	if (fseek(agi_file, 0, SEEK_SET) != 0)
		my_error("error seeking to start of agi file\n");
	
	printf ("The size of the AGI file is %ld bytes.\n", agi_size);
	if (  (agi_size < AGI_MIN) || (agi_size > AGI_MAX)  )
		printf("Warning: size is not between 34300 and 40600 bytes.\nMay not be a proper AGI file\n");
	
	
	// read the file
	agi_data = (u8*)malloc(agi_size);
	if (fread(agi_data, 1, agi_size, agi_file) != agi_size)
		my_error("incorrect number of bytes read from file.. that's an error btw\n");

	
	// check the signature
	if (  (agi_data[0] != 'M') ||  (agi_data[1] != 'Z')  )
		my_error("sig error.  agi file does not have the \"MZ\" signature.\nnot an exe file or the file is still compressed.\n");
	else
		printf("Signature is OK.  File is an EXE file.\n");
	
	
	// search
	printf("Searching for required bytes...  ");
	p1 = search_data(agi_data, agi_size, patch1);
	p2 = search_data(agi_data, agi_size, patch2);
	p3 = search_data(agi_data, agi_size, patch3);
	p4 = search_data(agi_data, agi_size, patch4);
	printf("\n");
	
	
	// patch
	printf("Patching the AGI file...\n");
	
	new_value[0] = (u8)(newstack & 0xFF);
	new_value[1] = (u8)((newstack >> 8) & 0xFF);
	
	*(p1) = new_value[0];
	*(p1+1) = new_value[1];
	*(p2) = new_value[0];
	*(p2+1) = new_value[1];
	*(p3) = new_value[0];
	*(p3+1) = new_value[1];
	*(p4) = new_value[0];
	*(p4+1) = new_value[1];

	if (fseek(agi_file, 0, SEEK_SET) != 0)
		my_error("error seeking to start of agi file\n");	
	if (fwrite(agi_data, 1, agi_size, agi_file) != agi_size)
		my_error("incorrect number of bytes written to file.  out of space?\n");

	
	
	// close the agi file
	// keep right at the end since we'll be making modifications
	printf("Closing AGI...\n");
	free(agi_data);
	if (fclose(agi_file) != 0)
		my_error("error closing the file.  contents may have not been saved.\n");
	
	return 0;
}